"""Module that provides wrappers around IOA Rule Groups and Custom IOA Rules."""
from __future__ import annotations
from datetime import datetime
from typing import Dict, List, Optional

from caracara.modules.custom_ioa.rule_types import RuleType


class IoaRuleGroup:
    """A class representing an IOA Rule Group.

    An IOA Rule Group stores Custom IOA Rules (represented by instances of `CustomIoaRule`), and can
    be assigned to any number of host groups.
    """

    # These fields should exist on all instances of this object
    name: str
    description: str
    platform: str
    rules: List[CustomIoaRule]
    rules_to_delete: List[CustomIoaRule]
    group: IoaRuleGroup

    # The fields below will only be populated if `exists_in_cloud()` returns `True`, with the
    # exception of `id` which should be initialised to `None` if this object does not exist on the
    # cloud
    id_: str
    comment: str
    committed_on: datetime
    created_by: str
    created_on: datetime
    customer_id: str
    deleted: bool
    enabled: str
    modified_by: str
    modified_on: datetime
    rule_ids: List[str]
    version: int

    def __init__(self, name: str, description: str, platform: str):
        """Create a new local IOA Rule Group.

        Rules can be added using the `add_rule` method, and this group can be created in the cloud
        using the `create_rule_group` request on the custom IOA module.

        Arguments
        ---------
        `name`: `str`
            The name of this rule group
        `description`: `str`
            The description of this rule group
        `platform`: `str`
            The platform that this rule group is for (typically `windows`, `linux` or `mac`)
        """
        self.name = name
        self.description = description
        self.platform = platform
        self.rules = []
        self.rules_to_delete = []
        self.group = None
        self.id_ = None

    def __repr__(self):
        """Return an unambiguous string representation of the IOA and its ID, platform and name."""
        return (
            f"<{self.__class__.__name__}(id_={repr(self.id_)}, version={repr(self.version)}, "
            f"platform={repr(self.platform)}, name={repr(self.name)}, ...)>"
        )

    @staticmethod
    def from_data_dict(data_dict: dict, rule_type_map: Dict[str, RuleType]) -> IoaRuleGroup:
        """Construct a rule group object from an instance of the `api.RuleGroupV1` API model.

        The API will return an api.RuleGroupV1 object. This function will parse the dictionary
        representation of this object and return an IoaRuleGroup. You must also provide a map
        of rule type IDs to their associated rule types.

        Arguments
        ---------
        `data_dict`: `dict`
            The data returned from the API conforming to `api.RuleGroupV1` in the Swagger doc
        `rule_type_map`: `Dict[str, RuleType]`
            A map from rule type ID to its corresponding rule type object

        Returns
        -------
        `IoaRuleGroup`: The resulting constructed group.
        """
        rule_group = IoaRuleGroup(
            name=data_dict["name"],
            description=data_dict["description"],
            platform=data_dict["platform"]
        )

        rule_group.comment = data_dict["comment"]
        rule_group.committed_on = data_dict["committed_on"]
        rule_group.created_by = data_dict["created_by"]
        rule_group.created_on = data_dict["created_on"]
        rule_group.customer_id = data_dict["customer_id"]
        rule_group.deleted = data_dict["deleted"]
        rule_group.enabled = data_dict["enabled"]
        rule_group.id_ = data_dict["id"]
        rule_group.modified_by = data_dict["modified_by"]
        rule_group.modified_on = data_dict["modified_on"]
        rule_group.rule_ids = data_dict["rule_ids"]
        rule_group.version = data_dict["version"]

        for raw_rule in data_dict.get("rules", []):
            # The following line might raise an index error if there's a rule type on this rule that
            # we don't know about. I don't catch this since I don't think it's likely to happen.
            rule_type = rule_type_map[raw_rule["ruletype_id"]]
            raw_rule["rulegroup_id"] = rule_group.id_  # API doesn't populate this field, so we do
            rule = CustomIoaRule.from_data_dict(data_dict=raw_rule, rule_type=rule_type)
            rule_group.rules.append(rule)

        return rule_group

    def add_rule(self, rule: CustomIoaRule):
        """Add a rule to a group.

        The rule specified is queued up for creation upon the next `update_rule_group` request.
        This method will fail if the provided rule is already associated with another object.

        Arguments
        ---------
        `rule`: `CustomIoaRule`
            The rule to add to this rule group
        """
        if rule.group is not None:
            raise Exception(
                "This rule has already been added to a group!"
            )
        rule.group = self
        self.rules.append(rule)

    def remove_rule(self, index_of_rule: int):
        """Remove a rule from a group.

        The rule specified is queued up for deletion upon the next `update_rule_group` request.

        Arguments
        ---------
        `index_of_rule`: `int`
            Index of the rule to delete within the `rules` member.
        """
        if index_of_rule >= len(self.rules) or index_of_rule < 0:
            raise Exception("Index of rule out of range!")

        removed_rule = self.rules.pop(index_of_rule)
        if removed_rule.exists_in_cloud():
            self.rules_to_delete.append(removed_rule)

    def validation(self):
        """Validate each rule to catch some errors before sending to the API.

        If there is an error an exception will be raised.
        """
        for rule in self.rules:
            rule.validation()

    def exists_in_cloud(self) -> bool:
        """Return if this object reflects an existing rule group in the cloud."""
        return self.id_ is not None

    def dump(self) -> dict:
        """Dump this rule group in conformance with api.RuleGroupV1 object model.

        This object model is defined in the CrowdStrike API Swagger document.
        """
        if not self.exists_in_cloud():
            raise Exception("This group does not exist in the cloud!")
        return {
            "customer_id": self.customer_id,
            "id": self.id_,
            "name": self.name,
            "description": self.description,
            "platform": self.platform,
            "enabled": self.enabled,
            "deleted": self.deleted,
            "rule_ids": [rule.instance_id for rule in self.rules],
            "rules": [rule.dump() for rule in self.rules],
            "version": self.version,
            "committed_on": self.committed_on,
            "created_on": self.created_on,
            "created_by": self.created_by,
            "modified_on": self.modified_on,
            "modified_by": self.modified_by,
            "comment": self.comment,
        }

    def dump_create(self, comment: str, verify: bool = True) -> dict:
        """Dump this rule group in conformance with the api.RuleGroupCreateRequestV1 object model.

        This object model is defined in the CrowdStrike API Swagger document. This function will
        also verify that this group has not been already created when `verify=True`.

        Arguments
        ---------
        `comment`: `str`
            The comment to associated with the creation action
        `verify`: `bool`, optional (default=`True`)
            Whether this method should verify whether the object has been created or not

        Returns
        -------
        `dict`: A dictionary structured in conformance with the api.RuleGroupCreateRequestV1 model.
        """
        if verify and self.exists_in_cloud():
            raise Exception("This group already exists in the cloud!")

        return {
            "name": self.name,
            "description": self.description,
            "platform": self.platform,
            "comment": comment,
        }

    def dump_update(self, comment: str, verify: bool = True) -> dict:
        """Dump this rule group in conformance with the api.RuleGroupModifyRequestV1 model.

        This object model is defined in the CrowdStrike API Swagger document. This function will
        also verify that this group has already been already created (and so can be updated)
        when `verify=True`.

        Arguments
        ---------
        `comment`: `str`
            The comment to associated with the update action.
        `verify`: `bool`, optional (default=`True`)
            Whether this method should verify whether the object has not been created or not

        Returns
        -------
        `dict`: A dictionary structured in conformance with the api.RuleGroupModifyRequestV1 model.
        """
        if verify and not self.exists_in_cloud():
            raise Exception("This group does not exist in the cloud!")

        return {
            "id": self.id_,
            "rulegroup_version": self.version,
            "name": self.name,
            "description": self.description,
            "enabled": self.enabled,
            "comment": comment,
        }

    def dump_rules_update(self, comment: str) -> dict:
        """Dump the current rules in a format conforming to the api.RuleUpdatesRequestV1 model.

        This object model is defined in the CrowdStrike API Swagger document.

        Arguments
        ---------
        `comment`: `str`
            The comment to associated with the update action.

        Returns
        -------
        `dict`: A dictionary structured in conformance with the api.RuleUpdatesRequestV1 model.
        """
        return {
            "comment": comment,
            "rule_updates": [rule.dump_update(group=self) for rule in self.rules],
            "rulegroup_version": self.version + 1,
            "rulegroup_id": self.id_,
        }


class CustomIoaRule:
    """A class representing a Custom IOA Rule.

    A Custom IOA Rule is associated with a single IOA Rule Group (represented by instances of
    `IoaRuleGroup`). So, a Custom IOA Rule is uniquely identified by a combination of its own ID and
    its group ID.

    A rule can be created with a provided `RuleType`. Right now you must query the rule types using
    the `describe_rule_types` request on the custom IOA module, and pick the one you want. In the
    future, we may provide a quicker method.

    Every rule needs to have its action/disposition set using `set_action`.

    Additionally, if a rule has options for a regex (known as an 'excludable' field), this has an
    include and an optional exclude regex that can be provided. These parameters can be set using
    the `set_excludable_field` method.

    Sometimes the rule may have an option to select a number of checkboxes in a set of values (known
    as a 'set' field). The options for this field can be set with `set_set_field`.

    The easiest way to find out the action/field names and possible values is to look at their names
    and options on the CrowdStrike Falcon web interface.
    """

    # These fields should exist on all instances of this object
    name: str
    description: str
    severity: str
    rule_type: RuleType
    fields: Dict[(str, str), dict]  # (name, type) -> raw dict

    # This field should exist if this object has been added to a group
    group: IoaRuleGroup

    # All these fields should exist if `exists_in_cloud()` returns True, with the exception of
    # instance_id which should exist and initialised to `None` if this object does not exist in the
    # cloud
    instance_id: str
    action_label: str
    comment: str
    committed_on: datetime
    created_by: str
    created_on: datetime
    customer_id: str
    deleted: bool
    disposition_id: str
    enabled: bool
    instance_version: int
    magic_cookie: int
    modified_by: str
    modified_on: datetime
    version_ids: List[str]
    pattern_id: str

    def __init__(
        self,
        name: str,
        description: str,
        severity: str,
        rule_type: RuleType,
    ):
        """Construct a local rule object.

        This rule needs to be customised using the `set_*` methods (see the class docstring for more
        information). After that, it can be added to a rule group using the `add_rule` method on an
        `IoaRuleGroup` object, and created in the cloud by either creating that group if it doesn't
        exist (with `create_rule_group`), or updating it if it already exists (with
        `update_rule_group`)

        Arguments
        ---------
        `name`: `str`
            The name of this rule
        `description`: `str`
            The description of this rule
        `severity`: `str`
            The severity of this rule (typically `informational`, `low`, `medium`, `high` or
            `critical`)
        `rule_type`: `RuleType`
            The type of this rule.
        """
        self.name = name
        self.description = description
        self.severity = severity
        self.rule_type = rule_type
        self.group = None
        self.instance_id = None

        self.fields = {}
        for field_type in rule_type.fields:
            field = field_type.to_concrete_field()
            self.fields[(field["name"], field["type"])] = field

    def __repr__(self):
        """Return an unambiguous string representation of the CustomIoaRule and its properties.

        This function will display the rule's group, ID, version, name and type.
        """
        return (
            f"<{self.__class__.__name__}(group={repr(self.group)}, "
            f"instance_id={repr(self.instance_id)}, instance_version={repr(self.instance_version)} "
            f"name={repr(self.name)}, ruletype={repr(self.rule_type)}, ...)>"
        )

    @staticmethod
    def from_data_dict(data_dict: dict, rule_type: RuleType) -> CustomIoaRule:
        """Construct a rule object from an instance of the `api.RuleV1` object.

        This object will conform to the `api.RuleV1` object model defined by the CrowdStrike API
        Swagger document, and will be returned from the API. You must also provide the rule's type
        for this function to succeed.

        Arguments
        ---------
        `data_dict`: `dict`
            The data returned from the API conforming to `api.RuleV1` in the Swagger doc
        `rule_type`: `RuleType`
            The type of this rule.

        Returns
        -------
        `CustomIoaRule`: The resulting constructed rule.
        """
        # note that this won't account for group
        rule = CustomIoaRule(
            name=data_dict["name"],
            description=data_dict["description"],
            severity=data_dict["pattern_severity"],
            rule_type=rule_type,
        )

        rule.action_label = data_dict["action_label"]
        rule.committed_on = data_dict["committed_on"]
        rule.created_by = data_dict["created_by"]
        rule.created_on = data_dict["created_on"]
        rule.customer_id = data_dict["customer_id"]
        rule.deleted = data_dict["deleted"]
        rule.description = data_dict["description"]
        rule.disposition_id = data_dict["disposition_id"]
        rule.enabled = data_dict["enabled"]
        rule.instance_id = data_dict["instance_id"]
        rule.instance_version = data_dict["instance_version"]
        rule.magic_cookie = data_dict["magic_cookie"]
        rule.modified_by = data_dict["modified_by"]
        rule.modified_on = data_dict["modified_on"]
        rule.version_ids = data_dict["version_ids"]
        rule.pattern_id = data_dict["pattern_id"]
        rule.comment = data_dict["comment"]

        rule.fields = {}
        for field_value in data_dict["field_values"]:
            rule.fields[(field_value["name"], field_value["type"])] = field_value

        return rule

    def validation(self):
        """Validate the IoaRuleGroup to catch errors before sending data to the API.

        This function will raise an exception if there are any issues with the data within this
        object, and contextual information will be provided alongside each exception.
        """
        # Check an action / disposition has been set
        if not hasattr(self, "disposition_id"):
            raise Exception("Rule has no action, make sure to set one with the set_action method")

        # Check that at least one excludable field is non-default
        regex_values = [
            value["value"]
            for field in self.fields.values() if field["type"] == "excludable"
            for value in field["values"]
        ]
        if len(regex_values) > 0 and all(value == ".*" for value in regex_values):
            raise Exception(
                "All excludable/regex fields set to the default of '.*' which the API will reject, "
                "set one to something else with the set_excludable_field method"
            )

    def exists_in_cloud(self) -> bool:
        """Return `True` if this rule object reflects an existing rule in the cloud."""
        return self.instance_id is not None

    def get_possible_actions(self) -> List[str]:
        """Return a list of possible actions to set using the `set_action` method."""
        return list(self.rule_type.disposition_map.values())

    def set_action(self, action: str or int):
        """Set the action/disposition for this rule.

        An action must be set for the API to accept this rule.

        Arguments
        ---------
        `action`: `str` or `int`
            The label of the action to set to, or the disposition ID of the action.
        """
        if isinstance(action, int):
            if action in self.rule_type.disposition_map.keys():
                self.disposition_id = action
                self.action_label = self.rule_type.disposition_map[action]
            else:
                raise Exception("Invalid action/disposition id!")
        else:
            matches = [id_ for (id_, label)
                       in self.rule_type.disposition_map.items() if label == action]
            if len(matches) > 0:
                self.disposition_id = matches[0]
                self.action_label = action
            else:
                raise Exception("Invalid action/disposition label!")

    def set_excludable_field(self, name_or_label: str, include: str, exclude: Optional[str] = None):
        """Set a field of type excludable with a matching name or label to the one provided.

        Fields of type 'excludable' concern regular expression matching. Each field has an include
        (defaulting to `.*`), and an optional exclude. At least one excludable field must be
        non-default for the API to accept this rule.

        Arguments
        ---------
        `name_or_label`: `str`
            The name or label of the field to set.
        `include`: `str`
            The include regex for this field.
        `exclude`: `Optional[str]`
            The exclude regex for this field (optional)
        """
        field = self.rule_type.get_field(name_or_label, "excludable")

        if field is None:
            raise Exception(
                f"Rule type {repr(self.rule_type.name)} has no fields with name or label "
                f"{repr(name_or_label)} and type \"excludable\""
            )

        values = [{
            "label": "include",
            "value": include,
        }]

        if exclude is not None:
            values.append({
                "label": "exclude",
                "value": exclude
            })

        self.fields[(field.name, "excludable")] = {
            "name": field.name,
            "label": field.label,
            "type": field.type,
            "values": values,
        }

    def get_set_field_options(self, name_or_label: str) -> List[str]:
        """Return the available options for a set field.

        Arguments
        ---------
        `name_or_label`: `str`
            The name or label of the set field to check the options of.
        """
        field = self.rule_type.get_field(name_or_label, "set")
        return [option.value for option in field.options]

    def set_set_field(self, name_or_label: str, selected_options: List[str]):
        """Set the options for a set field.

        A set field can be thought of as a list of checkboxes, where you can check a subset of a
        set of values. For a list of available options you cna use the `get_set_field_options`
        method. By default all options are checked.

        Arguments
        ---------
        `name_or_label`: `str`
            The name or label of the set field to set.
        `selected_options`: `List[str]`
            The list of the options selected for this field.
        """
        field = self.rule_type.get_field(name_or_label, "set")

        if field is None:
            raise Exception(
                f"Rule type {repr(self.rule_type.name)} has no fields with name or label "
                f"{repr(name_or_label)} and type \"excludable\""
            )

        values = []
        for value in selected_options:
            # Find all options that have 'label' or 'value' matching the selected option
            matching_options = [option for option in field.options if value in
                                [option.label, option.value]]

            if len(matching_options) == 0:
                raise Exception(
                    f"No option matching {repr(value)} in field {repr(field.name)} in rule with "
                    f"type \"set\""
                )

            option = matching_options[0]
            values.append(option.dump())

        self.fields[(field.name, "set")] = {
            "name": field.name,
            "label": field.label,
            "type": field.type,
            "values": values,
        }

    def dump(self) -> dict:
        """Dump this object in conformance with the api.RuleV1 model.

        This object model is defined in the CrowdStrike API Swagger document.
        """
        return {
            "customer_id": self.customer_id,
            "instance_id": self.instance_id,
            "name": self.name,
            "description": self.description,
            "pattern_id": self.pattern_id,
            "pattern_severity": self.severity,
            "disposition_id": self.disposition_id,
            "action_label": self.action_label,
            "ruletype_id": self.rule_type.id_,
            "ruletype_name": self.rule_type.name,
            "field_values": list(self.fields.values()),
            "enabled": self.enabled,
            "deleted": self.deleted,
            "instance_version": self.instance_version,
            "version_ids": self.version_ids,
            "magic_cookie": self.magic_cookie,
            "committed_on": self.committed_on,
            "created_on": self.created_on,
            "created_by": self.created_by,
            "modified_on": self.modified_on,
            "modified_by": self.modified_by,
            "comment": self.comment,
        }

    def dump_update(self, verify: bool = True) -> dict:
        """Dump this object in conformance with the api.RuleUpdateV1 model.

        This object model is defined in the CrowdStrike API Swagger document, excluding
        'rulegroup_version'. This function will check that this object reflects an existing
        (updatable) object in the cloud if verify=`True`.
        """
        if verify:
            if not self.exists_in_cloud():
                raise Exception("Can't update a rule that hasn't been created in the cloud")
            self.validation()

        return {
            "name": self.name,
            "description": self.description,
            "enabled": self.enabled,
            "instance_id": self.instance_id,
            "pattern_severity": self.severity,
            "disposition_id": self.disposition_id,
            "field_values": list(self.fields.values()),
        }

    def dump_create(self, comment: str, verify: bool = True) -> dict:
        """Dump this object in conformance with the `api.RuleCreateV1` model.

        This object model is defined in the CrowdStrike API Swagger document. The function will
        check that this object does not reflect an existing object in the cloud (and so can be
        safely created) if verify=`True`.
        """
        if verify:
            if self.exists_in_cloud():
                raise Exception("This rule already exists in the cloud!")
            self.validation()

        return {
            "name": self.name,
            "description": self.description,
            "pattern_severity": self.severity,
            "disposition_id": self.disposition_id,
            "field_values": list(self.fields.values()),
            "rulegroup_id": self.group.id_,
            "ruletype_id": self.rule_type.id_,
            "comment": comment,
        }



"""Caracara: Common functions and data imports."""
from caracara.common.common_filters import FILTER_ATTRIBUTES
from caracara.common.constants import FILTER_OPERATORS, SCROLL_BATCH_SIZE, DATA_BATCH_SIZE
from caracara.common.policy_wrapper import Policy
from caracara.filters.utils import IPChecker, RelativeTimestamp, ISOTimestampChecker
from caracara.common.meta import user_agent_string
from caracara.common.module import FalconApiModule

__all__ = [
    "FILTER_OPERATORS", "SCROLL_BATCH_SIZE", "DATA_BATCH_SIZE", "IPChecker",
    "RelativeTimestamp", "ISOTimestampChecker", "FILTER_ATTRIBUTES",
    "user_agent_string", "FalconApiModule", "Policy",
]



"""
Falcon Query Language (FQL) filter string builder module.

Allows for the automatic creation of FQL strings.
"""

__all__ = [
    "FalconFilter",
    "FalconFilterAttribute",
]

from caracara.filters.falcon_filter import FalconFilter
from caracara.filters.fql import FalconFilterAttribute



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['caracara',
 'caracara.common',
 'caracara.filters',
 'caracara.modules',
 'caracara.modules.custom_ioa',
 'caracara.modules.flight_control',
 'caracara.modules.hosts',
 'caracara.modules.prevention_policies',
 'caracara.modules.response_policies',
 'caracara.modules.rtr']

package_data = \
{'': ['*']}

install_requires = \
['crowdstrike-falconpy>=1.2.9,<2.0.0', 'py7zr>=0.20,<0.21']

entry_points = \
{'console_scripts': ['agent-versions = '
                     'examples.hosts.show_agent_versions:show_agent_versions',
                     'clear-queued-sessions = '
                     'examples.rtr.clear_queued_sessions:clear_queued_sessions',
                     'create-prevention-policy = '
                     'examples.prevention_policies.create_prevention_policy:create_prevention_policy',
                     'create-response-policy = '
                     'examples.response_policies.create_response_policy:create_response_policy',
                     'describe-prevention-policies = '
                     'examples.prevention_policies.describe_prevention_policies:describe_prevention_policies',
                     'describe-put-files = '
                     'examples.rtr.describe_put_files:describe_put_files',
                     'describe-queued-sessions = '
                     'examples.rtr.describe_queued_sessions:describe_queued_sessions',
                     'describe-response-policies = '
                     'examples.response_policies.describe_response_policies:describe_response_policies',
                     'describe-scripts = '
                     'examples.rtr.describe_scripts:describe_scripts',
                     'download-event-log = '
                     'examples.rtr.download_event_log:download_event_log',
                     'find-devices = examples.hosts.find_devices:find_devices',
                     'list-all-device-states = '
                     'examples.hosts.list_device_states:list_device_states',
                     'list-all-devices = '
                     'examples.hosts.list_all_devices:list_all_devices',
                     'list-all-group-member-ids = '
                     'examples.hosts.list_all_group_member_ids:list_all_group_member_ids',
                     'list-all-group-members = '
                     'examples.hosts.list_all_group_members:list_all_group_members',
                     'list-all-groups = '
                     'examples.hosts.list_all_groups:list_all_groups',
                     'list-device-address-changes = '
                     'examples.hosts.list_network_history:list_network_history',
                     'list-device-logins = '
                     'examples.hosts.list_login_history:list_login_history',
                     'list-hidden-devices = '
                     'examples.hosts.list_hidden_devices:list_hidden_devices',
                     'list-windows-devices = '
                     'examples.hosts.list_windows_devices:list_windows_devices',
                     'queue-command = examples.rtr.queue_command:queue_command',
                     'stale-sensors = '
                     'examples.hosts.find_stale_sensors:find_stale_sensors']}

setup_kwargs = {
    'name': 'caracara',
    'version': '0.2.1',
    'description': 'The CrowdStrike Falcon Developer Toolkit',
    'long_description': '![CrowdStrike Falcon](https://raw.githubusercontent.com/CrowdStrike/falconpy/main/docs/asset/cs-logo.png) [![Twitter URL](https://img.shields.io/twitter/url?label=Follow%20%40CrowdStrike&style=social&url=https%3A%2F%2Ftwitter.com%2FCrowdStrike)](https://twitter.com/CrowdStrike)<br/>\n\n# Caracara\n\n\n\n<!--\n![PyPI - Status](https://img.shields.io/pypi/status/caracara)\n[![Pylint](https://github.com/CrowdStrike/caracara/actions/workflows/pylint.yml/badge.svg)](https://github.com/CrowdStrike/caracara/actions/workflows/pylint.yml)\n[![Flake8](https://github.com/CrowdStrike/caracara/actions/workflows/flake8.yml/badge.svg)](https://github.com/CrowdStrike/caracara/actions/workflows/flake8.yml)\n[![Bandit](https://github.com/CrowdStrike/caracara/actions/workflows/bandit.yml/badge.svg)](https://github.com/CrowdStrike/caracara/actions/workflows/bandit.yml)\n[![CodeQL](https://github.com/CrowdStrike/caracara/actions/workflows/codeql.yml/badge.svg)](https://github.com/CrowdStrike/caracara/actions/workflows/codeql.yml)\n-->\n[![PyPI](https://img.shields.io/pypi/v/caracara)](https://pypi.org/project/caracara/)\n![OSS Lifecycle](https://img.shields.io/osslifecycle/CrowdStrike/caracara)\n\nA friendly wrapper to help you interact with the CrowdStrike Falcon API. Less code, less fuss, better performance, and full interoperability with [FalconPy](https://github.com/CrowdStrike/falconpy/).\n\n- [Features](#features)\n- [Installation](#installation-instructions)\n- [Basic Usage](#basic-usage-example)\n- [Examples](#examples-collection)\n- [Documentation](#documentation)\n- [Contributing](#contributing)\n\n## Features\n\nA few of the developer experience enhancements provided by the Caracara toolkit include:\n| Feature | Details |\n| :---  | :--- |\n| __Automatic pagination with concurrency__ | Caracara will handle all request pagination for you, so you do not have to think about things like batch sizes, batch tokens or parallelisation. Caracara will also multithread batch data retrieval requests where possible, dramatically reducing data retrieval times for large datasets such as host lists. |\n| __Friendly to your IDE (and you!)__ | Caracara is written with full support for IDE autocomplete in mind. We have tested autocomplete in Visual Studio Code and PyCharm, and will accept issues and patches for more IDE support where needed. Furthermore, all code, where possible, is written with type hints so you can be confident in parameters and return values. |\n| __Logging__ | Caracara is built with the in-box `logging` library provided with Python 3. Simply set up your logging handlers in your main code file, and Caracara will forward over `debug`, `info` and `error` logs as they are produced. Note that the `debug` logs are very verbose, and we recommend writing these outputs to a file as opposed to the console when retrieving large amounts of lightly filtered data. |\n| __Real Time Response (RTR) batch session abstraction__ | Caracara provides a rich interface to RTR session batching, allowing you to connect to as many hosts as possible. Want to download a specific file from every system in your Falcon tenant? Caracara will even extract it from the `.7z` container for you. |\n| __Rich and detailed sample code__ | Every module of Caracara comes bundled with executable, fully configurable code samples that address frequent use cases. All samples are built around a common structure allowing for code reuse and easy reading. Just add your API credentials to `config.yml`, and all samples will be ready to go. |\n| __Simple filter syntax__ | Caracara provides an object-orientated Falcon Query Language (FQL) generator. The `FalconFilter` object lets you specify filters such as `Hostname`, `OS` and `Role`, automatically converting them to valid FQL. Never write a FQL filter yourself again! |\n| __Single authentication point of entry__ | Authenticate once and have access to every module. |\n| __100% FalconPy compatibility__ | Caracara is built on FalconPy, and can even be configured with a FalconPy `OAuth2` object via the `auth_object` constructor parameter, allowing you to reuse FalconPy authentication objects across Caracara and FalconPy. Authenticate once with FalconPy, and access every feature of FalconPy and Caracara. |\n\n## Installation Instructions\n\nCaracara supports all major Python packaging solutions. Instructions for [Poetry](https://python-poetry.org) and [Pip](https://pypi.org/project/pip/) are provided below.\n\n<details>\n<summary><h3>Installing Caracara from PyPI using Poetry (Recommended!)</h3></summary>\n\n### Poetry: Installation\n\n```shell\npoetry add caracara\n```\n\n### Poetry: Upgrading\n\n```shell\npoetry update caracara\n```\n\n### Poetry: Removal\n\n```shell\npoetry remove caracara\n```\n</details>\n\n<details>\n<summary><h3>Installing Caracara from PyPI using Pip</h3></summary>\n\n### Pip: Installation\n\n```shell\npython3 -m pip install caracara\n```\n\n### Pip: Upgrading\n\n```shell\npython3 -m pip install caracara --upgrade\n```\n\n### Pip: Removal\n\n```shell\npython3 -m pip uninstall caracara\n```\n\n</details>\n\n## Basic Usage Examples\n\n```python\n"""List Windows devices.\n\nThis example will use the API credentials provided as keywords to list the\nIDs and hostnames of all systems within your Falcon tenant that run Windows.\n"""\n\nfrom caracara import Client\n\nclient = Client(\n    client_id="12345abcde",\n    client_secret="67890fghij",\n)\n\nfilters = client.FalconFilter()\nfilters.create_new_filter("OS", "Windows")\n\nresponse_data = client.hosts.describe_devices(filters)\nprint(f"Found {len(response_data)} devices running Windows")\n\nfor device_id, device_data in response_data.items():\n    hostname = device_data.get("hostname", "Unknown Hostname")\n    print(f"{device_id} - {hostname}")\n```\n\nYou can also leverage the built in context manager and environment variables.\n\n```python\n"""List stale sensors.\n\nThis example will use the API credentials set in the environment to list the\nhostnames and IDs of all systems within your Falcon tenant that have not checked\ninto your CrowdStrike tenant within the past 7 days.\n\nThis is determined based on the filter LastSeen less than or equal (LTE) to 7 days ago (-7d).\n"""\n\nfrom caracara import Client\n\n\nwith Client(client_id="${CLIENT_ID_ENV_VARIABLE}", client_secret="${CLIENT_SECRET_ENV_VARIABLE}") as client:\n    filters = client.FalconFilter()\n    filters.create_new_filter("LastSeen", "-7d", "LTE")\n    response_data = client.hosts.describe_devices(filters)\n\nprint(f"Found {len(response_data)} stale devices")\n\nfor device_id, device_data in response_data.items():\n    hostname = device_data.get("hostname", "Unknown Hostname")\n    print(f"{device_id} - {hostname}")\n```\n\n## Examples Collection\n\nEach API wrapper is provided alongside example code. Cloning or downloading/extracting this repository allows you to execute examples directly.\n\nUsing the examples collection requires that you install our Python packaging tool of choice, [Poetry](https://python-poetry.org). Please refer to the Poetry project\'s [installation guide](https://python-poetry.org/docs/#installation) if you do not yet have Poetry installed.\n\nOnce Poetry is installed, make sure you run `poetry install` within the root repository folder to set up the Python virtual environment.\n\nTo configure the examples, first copy `examples/config.example.yml` to `examples/config.yml`. Then, add your API credentials and example-specific settings to `examples/config.yml`. Once you have set up profiles for each Falcon tenant you want to test with, execute examples using one of the two options below.\n\n### Executing the Examples\n\nThere are two ways to use Poetry to execute the examples.\n\n<details>\n<summary><h4>Executing from a Poetry Shell</h4></summary>\n\nThe `poetry shell` command will enter you into the virtual environment. All future commands will run within the Caracara virtual environment using Python 3, until you run the `deactivate` command.\n\n```shell\npoetry shell\nexamples/get_devices/list_windows_devices.py\n```\n\n</details>\n\n<details>\n<summary><h4>Executing without Activating the Virtual Environment</h4></summary>\n\nIf you do not want to enter the Caracara virtual environment (e.g., because you are using your system\'s installation of Python for other purposes), you can use the `poetry run` command to temporarily invoke the virtual environment for one-off commands.\n\n```shell\npoetry run examples/get_devices/list_windows_devices.py\n```\n\nAll examples are also configured in the `pyproject.toml` file as scripts, allowing them to be executed simply.\n\n```shell\npoetry run stale-sensors\n```\n\n> To get a complete list of available examples, execute the command `util/list-examples.sh` from the root of the repository folder.\n\n</details>\n\n## Documentation\n\n__*Coming soon!*__\n\n## Contributing\n\nInterested in taking part in the development of the Caracara project? Start [here](CONTRIBUTING.md).\n\n## Why Caracara?\n\nSimple! We like birds at CrowdStrike, so what better bird to name a Python project after one that eats just about anything, including snakes :)\n',
    'author': 'CrowdStrike',
    'author_email': 'falconpy@crowdstrike.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Common constants to be shared throughout Caracara."""

FILTER_OPERATORS = {
    "EQUAL": '',
    "NOT": '!',
    "GREATER": '>',
    "GTE": '>=',
    "LESS": '<',
    "LTE": '<=',
}

# Batch size of data downloaded via a multi-threaded data pull
DATA_BATCH_SIZE = 500

# Default pagination limit
PAGINATION_LIMIT = 100

# Batch size of data downloaded from scrolled endpoints
SCROLL_BATCH_SIZE = 5000

# Host group batch size
HOST_GROUP_SCROLL_BATCH_SIZE = 100

# List of all platforms supported by Falcon
PLATFORMS = [
    "Linux",
    "Mac",
    "Windows",
]

#
DEFAULT_COMMENT = "This action was performed by the Caracara Python library."



"""Real Time Response (RTR) batch session abstraction class."""
import concurrent.futures
from dataclasses import dataclass
import logging

from datetime import datetime, timedelta
from functools import partial, wraps
from threading import current_thread
from typing import Dict, List

from falconpy import (
    RealTimeResponse,
    RealTimeResponseAdmin,
)

from caracara.common.batching import batch_data_pull_threads
from caracara.modules.rtr.constants import (
    MAX_BATCH_SESSION_HOSTS,
    MAX_BATCH_SESSION_THREADS,
    RTR_COMMANDS,
    SESSION_EXPIRY,
    SESSION_REFRESH_TIMEOUT,
)
from caracara.modules.rtr.get_file import GetFile


def _batch_session_required(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.batch_sessions is None:
            raise Exception("You must connect to hosts with the connect() function first")

        self.auto_refresh_sessions(self.default_timeout)

        return func(self, *args, **kwargs)
    return wrapper


class InnerRTRBatchSession:  # pylint: disable=too-few-public-methods
    """
    Container class representing a single batch session of up to 10,000 systems.

    Many of these may be used in parallel to support batches of > 10,000 systems.
    These are referred to colloquially as "batches of batches. In other words,
    a batch of batches an RTRBatchSession object, which itself contains
    a list of InnerRTRBatchSession objects.
    """

    batch_id: str = None
    devices: Dict = None
    expiry: datetime = None
    logger: logging.Logger = None

    def __init__(
        self,
        batch_id: str,
        devices: Dict,
        expiry: datetime,
        logger: logging.Logger
    ):
        """Configure an inner batch of RTR sessions."""
        self.batch_id = batch_id
        self.devices = devices
        self.expiry = expiry
        self.logger = logger.getChild(batch_id)

    def __str__(self):
        """Return a loggable string showing the RTR batch ID and system count."""
        return f"{self.batch_id}: {len(self.devices)} devices"


@dataclass
class BatchGetCmdRequest:
    """A batch get request targeted to a specific set of devices within a batch RTR session."""

    batch_get_cmd_req_id: str
    devices: Dict


def generic_rtr_worker(
    session: InnerRTRBatchSession,
    func: partial,
    logger: logging.Logger,
    device_ids: List[str] = None,
):
    """
    Execute a partial function against an RTR batch session.

    This is a generic function designed to be used by a ThreadPoolExecutor
    to execute a partial function against an RTR batch session.
    """
    thread_name = current_thread().name
    logger = logger.getChild(__name__)
    logger.info(
        "%s | Executing %s function against RTR batch: %s (args: %s; kwargs: %s)",
        thread_name, func.func.__name__, session.batch_id, func.args, func.keywords,
    )
    if device_ids:
        device_ids_in_batch = list(
            filter(lambda x: x in session.devices.keys(), device_ids)
        )
        func.keywords['optional_hosts'] = device_ids_in_batch
    response = func(batch_id=session.batch_id)['body']
    logger.debug("%s | %s", thread_name, response)
    return response


class RTRBatchSession:
    """
    Real Time Response (RTR) Batch Session.

    Enables interactions with one or more systems via the RTR API.
    One RTRBatchSession object can connect to an unlimited number of systems by abstracting
    away a list of InnerRTRBatchSession objects.
    Each InnerRTRBatchSession object behind the scenes can communicate with up to 10,000
    individual systems (known as a "batch"). An RTRBatchSession can then multiplex these
    to command "batches of batches" of systems, thus removing the cloud-enforced connection
    limit.
    """

    admin_api: RealTimeResponseAdmin = None
    api: RealTimeResponse = None
    batch_sessions: List[InnerRTRBatchSession] = None
    default_timeout = 30
    logger = logging.getLogger(__name__)

    @_batch_session_required
    def device_ids(self) -> List[str]:
        """Return a list of device IDs from all inner batch sessions."""
        return [x.devices.keys() for x in self.batch_sessions]

    def connect(
        self,
        device_ids: List[str],
        queueing: bool = False,
        timeout: int = default_timeout,
    ):
        """
        Establish a connection to one or more hosts.

        This function must be executed before any other commands can be run.
        """
        self.logger.info("Establishing an RTR batch session with %d systems", len(device_ids))
        self.logger.debug(device_ids)

        batches = []
        for i in range(0, len(device_ids), MAX_BATCH_SESSION_HOSTS):
            batches.append(device_ids[i:i+MAX_BATCH_SESSION_HOSTS])
        self.logger.info("Divided up devices into %d batches", len(batches))

        def worker(batch_device_ids: List[str], batch_func: partial):
            thread_name = current_thread().name
            self.logger.info(
                "%s | Batch worker started with a list of %d devices",
                thread_name, len(batch_device_ids),
            )
            response = batch_func(host_ids=batch_device_ids)['body']
            resources = response['resources']
            self.logger.info("%s | Connected to %s systems", thread_name, len(resources))
            self.logger.debug("%s | %s", thread_name, response)
            batch_data = InnerRTRBatchSession(
                batch_id=response['batch_id'],
                devices=resources,
                expiry=datetime.now() + timedelta(seconds=SESSION_EXPIRY),
                logger=self.logger,
            )
            return batch_data

        batch_func = partial(
            self.api.batch_init_sessions,
            queue_offline=queueing,
            timeout=timeout,
            timeout_duration=f'{timeout}s',
        )

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=MAX_BATCH_SESSION_THREADS
        ) as executor:
            completed = executor.map(worker, batches, [batch_func])

        self.batch_sessions = []
        for complete in completed:
            self.logger.info("Completed a batch of RTR connections")
            self.batch_sessions.append(complete)

        device_count = sum(len(d.devices) for d in self.batch_sessions)
        self.logger.info("Connected to %d devices", device_count)
        self.logger.debug(self.batch_sessions)

    @_batch_session_required
    def disconnect(self):
        """Disconnect the RTR batch session."""
        for session in self.batch_sessions:
            self.logger.info("Disconnecting batch RTR session with ID %s", session.batch_id)
            self.api.delete_session(session.batch_id)

    @_batch_session_required
    def get(
        self,
        file_path: str,
        device_ids: List[str] = None,
        timeout: int = default_timeout,
    ) -> List[BatchGetCmdRequest]:
        """
        Upload a file to the Falcon cloud from the path specified.

        This function will execute the GET command against every system within every inner batch
        session. This allows you to retrieve a file at the same path from every connected system.
        Returns a tuple containing the batch get request ID and the responses
        """
        self.logger.info("Using a batch GET to retrieve the file at path %s", file_path)

        partial_get_func = partial(
            self.api.batch_get_command,
            file_path=file_path,
            timeout=timeout,
            timeout_duration=f'{timeout}s',
        )
        partial_worker = partial(generic_rtr_worker, logger=self.logger, device_ids=device_ids)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=MAX_BATCH_SESSION_THREADS
        ) as executor:
            completed: List[Dict] = executor.map(
                partial_worker,
                self.batch_sessions,
                [partial_get_func],
            )

        batch_get_cmd_reqs: List[BatchGetCmdRequest] = []
        for complete in completed:
            self.logger.info("Executed commands on a batch of %d hosts", len(complete))
            batch_get_cmd_req = BatchGetCmdRequest(
                batch_get_cmd_req_id=complete['batch_get_cmd_req_id'],
                devices=complete['combined']['resources'],
            )
            batch_get_cmd_reqs.append(batch_get_cmd_req)

        return batch_get_cmd_reqs

    def get_status(
        self,
        batch_get_cmd_reqs: List[BatchGetCmdRequest],
        timeout: int = default_timeout
    ) -> List[GetFile]:
        """
        Get a list of successful file uploads based on a list of batch get command requests.

        Each returned GetFile object will represent a file that has been successfully uploaded as a
        result of a batch get command.
        """
        self.logger.info("Checking the status of %d batch get requests", len(batch_get_cmd_reqs))

        def worker(batch_get_cmd_req: BatchGetCmdRequest, func: partial) -> List[GetFile]:
            thread_name = current_thread().name
            logger = self.logger.getChild(__name__)
            logger.info(
                "%s | Getting the status of batch get request %s",
                thread_name, batch_get_cmd_req.batch_get_cmd_req_id,
            )
            response = func(batch_get_cmd_req_id=batch_get_cmd_req.batch_get_cmd_req_id)['body']
            logger.debug("%s | %s", thread_name, response)

            resources: Dict = response['resources']
            get_files: List[GetFile] = []
            for device_id, get_data in resources.items():
                get_file = GetFile(
                    device_id=device_id,
                    filename=get_data['name'],
                    session_id=get_data['session_id'],
                    sha256=get_data['sha256'],
                    size=get_data['size'],
                    batch_session=self,
                )
                get_files.append(get_file)

            return get_files

        threads = batch_data_pull_threads()
        partial_func = partial(
            self.api.batch_get_command_status,
            timeout=timeout,
            timeout_duration=f'{timeout}s',
        )
        with concurrent.futures.ThreadPoolExecutor(max_workers=threads) as executor:
            completed = executor.map(worker, batch_get_cmd_reqs, [partial_func])

        all_get_files: List[GetFile] = []

        for complete in completed:
            all_get_files.extend(complete)

        return all_get_files

    def get_status_by_req_id(
        self,
        batch_get_cmd_req_id: str,
        timeout: int = default_timeout,
    ) -> List[GetFile]:
        """Get a list of successful file uploads based on an individual batch request ID."""
        self.logger.info("Checking the status of a batch get with ID %s", batch_get_cmd_req_id)
        response = self.api.batch_get_command_status(
            batch_get_cmd_req_id=batch_get_cmd_req_id,
            timeout=timeout,
            timeout_duration=f'{timeout}s',
        )['body']
        self.logger.debug(response)
        resources: List[Dict] = response['resources']
        self.logger.info("Batch GET has retrieved %d files so far", len(resources))
        self.logger.debug(resources)

        if not resources:
            return []

        get_files: List[GetFile] = []
        for device_id in resources.keys():
            get_file = GetFile(
                device_id=device_id,
                filename=resources[device_id]['name'],
                session_id=resources[device_id]['session_id'],
                sha256=resources[device_id]['sha256'],
                size=resources[device_id]['size'],
                batch_session=self,
            )
            get_files.append(get_file)

        return get_files

    def refresh_sessions(self, timeout: int = default_timeout):
        """Refresh a batch RTR session, resetting the timeout back to 10 minutes."""
        # Check whether we are connected to a session in this function rather than via the
        # _batch_session_required decorator to avoid an infinite loop
        if self.batch_sessions is None:
            raise Exception("You must connect to hosts with the connect() function first")

        self.logger.info("Refreshing batch RTR session")

        def worker(session: InnerRTRBatchSession, func: partial):
            thread_name = current_thread().name
            self.logger.info("%s | Refreshing batch session %s", thread_name, session.batch_id)
            func(batch_id=session.batch_id)
            return session.batch_id

        batch_func = partial(
            self.api.batch_refresh_sessions,
            timeout=timeout,
            timeout_duration=f'{timeout}s',
        )

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=MAX_BATCH_SESSION_THREADS
        ) as executor:
            completed = executor.map(worker, self.batch_sessions, [batch_func])

        for complete in completed:
            self.logger.info("Refreshed session %s", complete)

    @_batch_session_required
    def run_generic_command(
        self,
        command_string: str,
        device_ids: List[str] = None,
        timeout: int = default_timeout,
    ) -> Dict:
        """Execute an RTR command against all systems in the batch session."""
        base_command = command_string.split(' ')[0]
        if base_command not in RTR_COMMANDS:
            raise Exception(f"{base_command} is not a valid RTR command")

        self.logger.info("Executing a command via RTR: %s", command_string)

        # Permissions level will either be a string (e.g. admin), or
        # a dictionary of further permission levels based on the first
        # command parameter (such as 'query' for the reg command, or
        # -Raw for the runscript command)
        permissions_level = RTR_COMMANDS[base_command]
        if isinstance(permissions_level, dict):
            command_parameter_1 = command_string.split(' ')[1].split('=')[0]
            if command_parameter_1 in permissions_level:
                permissions_level = permissions_level[command_parameter_1]
            else:
                permissions_level = permissions_level['_default']

        if permissions_level == "admin":
            cmd_func = self.admin_api.batch_admin_command
        elif permissions_level == "active_responder":
            cmd_func = self.api.batch_active_responder_command
        elif permissions_level == "read_only":
            cmd_func = self.api.batch_command
        else:
            raise Exception(f"{base_command} does not have a valid permissions level")

        partial_cmd_func = partial(
            cmd_func,
            base_command=base_command,
            command_string=command_string,
            timeout=timeout,
            timeout_duration=f'{timeout}s',
        )
        partial_worker = partial(generic_rtr_worker, logger=self.logger, device_ids=device_ids)
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=MAX_BATCH_SESSION_THREADS
        ) as executor:
            completed: List[Dict] = executor.map(
                partial_worker,
                self.batch_sessions,
                [partial_cmd_func],
            )

        all_responses: Dict = {}
        for complete in completed:
            self.logger.info("Executed commands on a batch of %d hosts", len(complete))
            all_responses.update(complete['combined']['resources'])

        return all_responses

    @_batch_session_required
    def run_raw_script(
        self,
        script_text: str,
        script_timeout: int = 60,
        device_ids: List[str] = None,
        timeout: int = None,
    ) -> Dict:
        """Execute a raw RTR script on all systems in the batch session."""
        if not timeout:
            timeout = script_timeout + 10

        self.logger.info(
            "Running a raw script via RTR. Timeout: %d; script timeout: %d",
            timeout, script_timeout,
        )

        command_string = f"runscript -Raw=```{script_text}``` -Timeout={script_timeout}"
        self.logger.debug(command_string)
        resources = self.run_generic_command(command_string, device_ids, timeout)
        return resources

    def auto_refresh_sessions(self, timeout: int = default_timeout):
        """
        Automatically refresh every RTR batch session if they are going to expire soon.

        This function gets called by every function that makes an API request
        to the RTR API. This is done to attempt to ensure that we do not lose
        an active RTR session due to it expiring.
        We will refresh an RTR session before making a call to it if there are
        fewer than three minutes remaining.
        """
        session_near_expiry = False
        for session in self.batch_sessions:
            remaining_secs = (session.expiry - datetime.now()).total_seconds()
            self.logger.debug("RTR session %s has %ds remaining", session.batch_id, remaining_secs)
            if remaining_secs < SESSION_REFRESH_TIMEOUT:
                session_near_expiry = True
        if session_near_expiry:
            self.refresh_sessions(timeout)

    def __init__(  # pylint: disable=too-many-arguments
        self,
        rtr_api: RealTimeResponse,
        rtr_admin_api: RealTimeResponseAdmin,
        default_timeout: int = 30,
        device_ids: List[str] = None,
        queueing: bool = False,
        timeout: int = default_timeout,
    ):
        """
        Create an RTRBatchSession object, representing a batch of remote systems.

        If you pass a list of Device Agent IDs (AIDs), this constructor will automatically
        call the inner connect() function, thus establishing as many batch sessions as required
        to establish remote connections to these systems.
        """
        self.api = rtr_api
        self.admin_api = rtr_admin_api
        self.default_timeout = default_timeout

        # Provides the option to connect straight away on class init
        if device_ids is not None:
            self.connect(
                device_ids=device_ids,
                queueing=queueing,
                timeout=timeout,
            )

    def __enter__(self):
        """Treat an RTR batch session as a context manager."""
        if not self.batch_sessions:
            raise Exception(
                "You should pass device_ids to the constructor to automatically "
                "connect and the context manager functionality"
            )

    def __exit__(self, *args, **kwargs):
        """Disconnect from all systems when leaving the context manager."""
        self.disconnect()



"""Module defining wrappers around custom IOA rule types."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Dict, List, Optional


@dataclass
class RuleType:
    """A dataclass representing a custom IOA rule type."""

    id_: str
    name: str
    long_desc: str
    platform: str
    disposition_map: Dict[int, str]
    fields: List[RuleTypeField]
    released: bool
    channel: int

    def __repr__(self) -> str:
        """Return a string representation of the RuleType and its ID, platform and name."""
        return (
            f"<RuleType(id_={repr(self.id_)}, name={repr(self.name)}, "
            f"platform={repr(self.platform)}, ...)>"
        )

    @staticmethod
    def from_data_dict(data_dict: dict) -> RuleType:
        """Create an instance of this class using a dict conforming to the `api.RuleTypeV1` model.

        This object is mostly returned from the API. The Swagger model for api.RuleTypeV1 (as of
        writing) is inconsistent with what is returned from the cloud. If you are curious as to
        the actual structure, read the `RuleTypeField.dump` docstring.

        Arguments
        ---------
        `data_dict`: `dict`
            The dictionary conforming to api.RuleTypeV1 (mostly)

        Returns
        -------
        `RuleType`: the newly constructed rule type object
        """
        disposition_map = {}
        for mapping in data_dict["disposition_map"]:
            disposition_map[mapping["id"]] = mapping["label"]

        fields = []
        for raw_field in data_dict["fields"]:
            fields.append(RuleTypeField.from_data_dict(raw_field))

        rule_type = RuleType(
            id_=data_dict["id"],
            name=data_dict["name"],
            long_desc=data_dict["long_desc"],
            platform=data_dict["platform"],
            disposition_map=disposition_map,
            fields=fields,
            released=data_dict["released"],
            channel=data_dict["channel"],
        )

        return rule_type

    def get_field(
        self, field_name_or_label: str, field_type: Optional[str] = None
    ) -> Optional[RuleTypeField]:
        """Get a reference to a field via its its name or label and optionally its type.

        Arguments
        ---------
        `field_name_or_label`: `str`
            The name or label of the field to get
        `field_type`: `Optional[str]`
            The type of the field to get (optional)

        Returns
        -------
        `Optional[RuleTypeField]`: The rule type field if one can be found, `None` otherwise.
        """
        for field in self.fields:
            if (field_name_or_label in [field.name, field.label]
                    and (field_type is None or field.type == field_type)):
                return field
        return None

    def dump(self) -> dict:
        """Dump a dictionary representing this rule conforming to the api.RuleTypeV1 model.

        This object model is defined in the CrowdStrike API Swagger document.
        """
        dumped = {
            "id": self.id_,
            "name": self.name,
            "long_desc": self.long_desc,
            "platform": self.platform,
            "disposition_map": [
                {"id": k, "label": v} for k, v in self.disposition_map.items()
            ],
            "fields": [field.dump() for field in self.fields],
            "released": self.released,
            "channel": self.channel,
        }

        return dumped


@dataclass
class RuleTypeField:
    """A wrapper around a rule type's field.

    In the Swagger doc, this should be `domain.Field` according to `api.RuleTypeV1`, but in reality
    the API also provides `type` and options `fields`.
    """

    label: str
    name: str
    type: str
    options: List[RuleTypeFieldOption]

    @staticmethod
    def from_data_dict(data_dict: dict) -> RuleTypeField:
        """Construct an instance of this class from a dict almost conforming to `domain.Field`.

        For more information on `domain.Field`, see the class docstring.
        """
        options = []
        for raw_option in data_dict["options"]:
            options.append(RuleTypeFieldOption.from_data_dict(raw_option))

        field = RuleTypeField(
            label=data_dict["label"],
            name=data_dict["name"],
            type=data_dict["type"],
            options=options,
        )

        return field

    def dump(self) -> dict:
        """Dump this rule type field as a dict almost conforming to `domain.Field`.

        For more information on `domain.Field`, see the class docstring.
        """
        return {
            "label": self.label,
            "name": self.name,
            "type": self.type,
            "options": [option.dump() for option in self.options]
        }

    def to_concrete_field(self) -> dict:
        """Convert this spec for a field into an actual field that can be used in rules.

        This function will set default values where needed.
        """
        field = {
            "label": self.label,
            "name": self.name,
            "type": self.type,
        }

        if self.type == "excludable":
            field["values"] = [{"label": "include", "value": ".*"}]
        elif self.type == "set":
            field["values"] = [option.dump() for option in self.options]
        else:
            raise Exception(f"Unknown rule field type {repr(self.type)}")

        return field


@dataclass
class RuleTypeFieldOption:
    """A dataclass representing an option within a rule type field."""

    label: str
    value: str

    @staticmethod
    def from_data_dict(data_dict: dict) -> RuleTypeFieldOption:
        """Construct an instance of this object from an option returned by the API."""
        return RuleTypeFieldOption(
            label=data_dict["label"],
            value=data_dict["value"],
        )

    def dump(self):
        """Dump this option as a dict to be used in API calls."""
        return {"label": self.label, "value": self.value}



"""Caracara exceptions."""
from typing import Dict, List


class BaseCaracaraError(Exception):
    """Base exception class from which all other exceptions inherit."""


class GenericAPIError(BaseCaracaraError):
    """A generic error from the API."""

    def __init__(self, errors: List[Dict] = None):
        """Construct an instance of the GenericAPIError class."""
        self.errors = [{
            "code": 500,
            "message": "An unexpected error has occurred"
        }]
        if errors:
            self.errors = errors
        super().__init__(self.errors)

    def __str__(self):
        """Return all errors as a comma-delimited list."""
        return ", ".join([f"[{x['code']}] {x['message']}" for x in self.errors])

    def __int__(self):
        """Return the first error code as an integer."""
        return int(self.errors[0]["code"])

    def __float__(self):
        """Return the first error code as a float."""
        return float(self.errors[0]["code"])


class MustProvideFilterOrID(GenericAPIError):
    """You must provide either a Falcon Filter or an ID to use this method."""

    def __init__(self):
        """Construct an instance of the MustProvideFilterOrID class."""
        self.errors = [{
            "code": 500,
            "message": "You must provide either a Falcon Filter or an ID to use this method"
        }]
        super().__init__(self.errors)


class MustProvideFilter(GenericAPIError):
    """You must provide a Falcon Filter in order to use this method."""

    def __init__(self):
        """Construct an instance of the MustProvideFilter class."""
        self.errors = [{
            "code": 500,
            "message": "You must provide a Falcon Filter in order to use this method"
        }]
        super().__init__(self.errors)


class HostGroupNotFound(GenericAPIError):
    """The host group you specified is not found."""

    def __init__(self):
        """Construct an instance of the HostGroupNotFound class."""
        self.errors = [{
            "code": 404,
            "message": "The Falcon Filter you provided returned no Host Group matches"
        }]
        super().__init__(self.errors)


class DeviceNotFound(GenericAPIError):
    """The device you specified is not found."""

    def __init__(self):
        """Construct an instance of the DeviceNotFound class."""
        self.errors = [{
            "code": 404,
            "message": "The Falcon Filter you provided returned no device matches"
        }]
        super().__init__(self.errors)


class MissingArgument(GenericAPIError):
    """A required argument was not provided."""

    def __init__(self, arg_name: str = None):
        """Construct an instance of the MissingArgument class."""
        if not arg_name:
            arg_name = "A required argument"
        else:
            arg_name = f"The required argument {arg_name}"
        arg_str = f"{arg_name} was not provided"

        self.errors = [{
            "code": 500,
            "message": arg_str
        }]
        super().__init__(self.errors)


class MissingArguments(GenericAPIError):
    """A required argument was not provided."""

    def __init__(self, arg_names: List[str] = None):
        """Construct an instance of the MissingArguments class."""
        if not arg_names:
            arg_names = "Required arguments"
        else:
            arg_names = f"The required arguments {', '.join(arg_names)}"
        arg_str = f"{arg_names} were not provided"

        self.errors = [{
            "code": 500,
            "message": arg_str
        }]
        super().__init__(self.errors)



"""Caracara Real Time Response (RTR) module."""
__all__ = [
    'BatchGetCmdRequest',
    'FILTER_ATTRIBUTES',
    'GetFile',
    'InnerRTRBatchSession',
    'RTRApiModule',
    'RTRBatchSession',
]

from caracara.modules.rtr.batch_session import (
    BatchGetCmdRequest,
    InnerRTRBatchSession,
    RTRBatchSession,
)
from caracara.modules.rtr.get_file import GetFile
from caracara.modules.rtr.rtr import RTRApiModule
from caracara.modules.rtr.rtr_filters import FILTER_ATTRIBUTES



r"""Caracara Real Time Response (RTR) module.

 ______              _    _______ _                ______
(_____ \            | |  (_______|_)              (_____ \
 _____) )_____ _____| |      _    _ ____  _____    _____) )_____  ___ ____   ___  ____   ___ _____
|  __  /| ___ (____ | |     | |  | |    \| ___ |  |  __  /| ___ |/___)  _ \ / _ \|  _ \ /___) ___ |
| |  \ \| ____/ ___ | |     | |  | | | | | ____|  | |  \ \| ____|___ | |_| | |_| | | | |___ | ____|
|_|   |_|_____)_____|\_)    |_|  |_|_|_|_|_____)  |_|   |_|_____|___/|  __/ \___/|_| |_(___/|_____)
                                                                     |_|
"""
import os

from datetime import datetime
from functools import partial
from typing import Dict, List

from falconpy import (
    OAuth2,
    RealTimeResponse,
    RealTimeResponseAdmin,
)

from caracara.common.batching import batch_get_data
from caracara.common.module import FalconApiModule
from caracara.common.pagination import all_pages_numbered_offset_parallel
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string
from caracara.modules.rtr.batch_session import RTRBatchSession


class RTRApiModule(FalconApiModule):
    """Real Time Response (RTR) API Module."""

    name = "RTR API Module"
    help = "Interface with hosts via the Real Time Response (RTR) API"

    default_timeout = 30

    def __init__(self, api_authentication: OAuth2):
        """Create an RTR module object and configure it with a FalconPy OAuth2 object."""
        super().__init__(api_authentication)
        self.rtr_api = RealTimeResponse(auth_object=self.api_authentication)
        self.rtr_admin_api = RealTimeResponseAdmin(auth_object=self.api_authentication)

    def batch_session(self) -> RTRBatchSession:
        """
        Create a new RTR batch session abstraction object.

        Once created, use the connect() function to establish a connection to a list
        of Falcon device IDs.

        Returns
        -------
        RTRBatchSession: a batch RTR session object that can connect to many systems at once
        """
        rtr_batch_session = RTRBatchSession(
            rtr_api=self.rtr_api,
            rtr_admin_api=self.rtr_admin_api,
        )
        return rtr_batch_session

    @filter_string
    def _search_sessions(self, filters: str or FalconFilter = None):
        """
        Get RTR Session IDs based on filters.

        Arguments
        ---------
        filters: FalconFilter or str, optional
            Filters to apply to the RTR session search.

        Returns
        -------
        List[str]: A list of all session IDs discovered.
        """
        self.logger.info("Searching for RTR sessions based on filter string: %s", filters)
        func = partial(self.rtr_api.list_all_sessions, filter=filters)
        session_ids = all_pages_numbered_offset_parallel(func=func, logger=self.logger)
        return session_ids

    def _get_queued_session_ids(self) -> List[str]:
        """
        Get a list of queued RTR Session IDs.

        Returns
        -------
        List[str]: A list of IDs of all queued RTR sessions discovered.
        """
        self.logger.info("Searching for queued RTR sessions")
        filter_str = "offline_queued: 1+deleted_at: null"
        session_ids = self._search_sessions(filters=filter_str)
        return session_ids

    def describe_queued_sessions(self) -> Dict:
        """
        Show the contents of all queued RTR sessions.

        Returns
        -------
        Dict: A dictionary containing all queued RTR sessions discoverd.
        """
        self.logger.info("Describing all queued RTR sessions")
        session_ids = self._get_queued_session_ids()
        session_data = batch_get_data(session_ids, self.rtr_api.list_sessions)
        return session_data

    def delete_queued_session(self, session_id: str):
        """
        Delete a queued session by ID.

        Arguments
        ---------
        session_id: str
            ID of the queued RTR session to delete.

        Returns
        -------
        None
        """
        self.logger.info("Deleting queued session with ID %s", session_id)
        self.rtr_api.delete_session(session_id)

    def delete_queued_session_command(self, session_id: str, cloud_request_id: str):
        """Delete a specific command within a queued session."""
        self.logger.info(
            "Deleting command with ID %s from queued session %s",
            cloud_request_id, session_id,
        )
        self.rtr_api.delete_queued_session(
            session_id=session_id,
            cloud_request_id=cloud_request_id
        )

    def clear_queued_sessions(self):
        """
        Clear all sessions from the RTR queue.

        This is a shortcut function provided for a common operation, and can be done
        manually without this function.

        Returns
        -------
        None
        """
        self.logger.info("Clearing all queued RTR sessions")
        session_ids = self._get_queued_session_ids()
        for session_id in session_ids:
            self.delete_queued_session(session_id)

    @filter_string
    def describe_put_files(self, filters: str or FalconFilter = None) -> Dict:
        """
        Query RTR PUT files and return a list. Falcon (FQL) filters are supported.

        Arguments
        ---------
        filters: FalconFilter or str, optional
            Filters to apply to the PUT file search.

        Returns
        -------
        Dict: a dictionary containing information about all the PUT files discovered.
        """
        self.logger.info("Querying RTR put files using the filter string %s", filters)
        func_ids = partial(self.rtr_admin_api.list_put_files, filter=filters)
        put_file_ids = all_pages_numbered_offset_parallel(
            func=func_ids,
            logger=self.logger,
        )
        self.logger.info("Retreived %d PUT file IDs", len(put_file_ids))
        self.logger.debug(put_file_ids)

        put_file_data = batch_get_data(put_file_ids, self.rtr_admin_api.get_put_files)
        return put_file_data

    def create_put_file(self, file_path: str, name: str = None, description: str = None):
        """
        Create a PUT file within the Falcon cloud for later use.

        Arguments
        ---------
        file_path: str
            Path to the file on disk to be uploaded to the Falcon Cloud
        name: str, optional
            Rename the file when uploaded to the Falcon Cloud
        description: str, optional
            Description text to add to the file.
            Defaults to "File uploaded via Caracara at <current UTC timestamp>"

        Returns
        -------
        None
        """
        if not os.path.isfile(file_path):
            self.logger.info("%s is not a valid file", file_path)
            raise Exception("You must provide the path to a valid file on disk")

        # Default to using the original filename in the case that the user doesn't
        # specify an alternative
        if name is None:
            name = os.path.basename(file_path)

        if description is None:
            timestamp_str = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
            description = f"File uploaded via Caracara at {timestamp_str}"

        self.logger.info(
            "Uploading %s to Falcon with name %s and description %s",
            file_path, name, description,
        )

        with open(file_path, 'rb') as put_file:
            file_contents = put_file.read()

        self.rtr_admin_api.create_put_files(
            name=name,
            description=description,
            files=[(name, (name, file_contents, 'application/script'))]
        )

    def delete_put_file(self, put_file_id: str):
        """
        Delete a PUT file.

        Arguments
        ---------
        put_file_id: str
            ID of the PUT file to delete from the Falcon Cloud

        Returns
        -------
        None
        """
        self.logger.info("Deleting PUT file with ID %s", put_file_id)
        self.rtr_admin_api.delete_put_files(put_file_id)

    @filter_string
    def describe_scripts(self, filters: str or FalconFilter = None) -> Dict:
        """
        Query RTR scripts and return a list. Falcon (FQL) filters are supported.

        Arguments
        ---------
        filters: str or FalconFilter, optional
            Filters to apply to the script search

        Returns
        -------
        Dict: a dictionary containing a list of every script uploaded to the Falcon Cloud,
            including its contents
        """
        self.logger.info("Querying RTR scripts using the filter string %s", filters)
        func_ids = partial(self.rtr_admin_api.list_scripts, filter=filters)
        script_ids = all_pages_numbered_offset_parallel(
            func=func_ids,
            logger=self.logger,
        )
        self.logger.info("Retreived %d script file IDs", len(script_ids))
        self.logger.debug(script_ids)

        script_data = batch_get_data(script_ids, self.rtr_admin_api.get_scripts)
        return script_data



"""Caracara Hosts Module: host hiding functions.

In order to avoid the main hosts.py file getting too unwieldly, this file contains
the implementations of the functions required to hide and unhide hosts from the
Falcon UI.
"""
# Disable the protected access rule because this file is an extension of the class in hosts.py.
# pylint: disable=protected-access
from __future__ import annotations
from functools import partial
from typing import (
    Dict,
    List,
    TYPE_CHECKING,
)

from caracara.common.batching import batch_get_data
from caracara.common.constants import SCROLL_BATCH_SIZE
from caracara.common.exceptions import (
    DeviceNotFound,
    MustProvideFilter,
)
from caracara.common.pagination import all_pages_numbered_offset_parallel
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string

if TYPE_CHECKING:
    from caracara.modules.hosts import HostsApiModule


@filter_string
def describe_hidden_devices(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict:
    """Return a dictionary containing details for every hidden device in your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the hidden device search.

    Returns
    -------
    dict: A dictionary containing details for every hidden device discovered.
    """
    self.logger.info("Describing hidden devices based on the filter string %s", filters)
    device_ids = self.get_hidden_ids(filters)
    device_data = batch_get_data(device_ids, self.hosts_api.get_device_details)

    return device_data


@filter_string
def get_hidden_ids(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> List[str]:
    """Return a list of IDs (string) for every hidden device in your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the device search.

    Returns
    -------
    List[str]: A list of all device IDs discovered.
    """
    self.logger.info("Getting the IDs of all hidden devices using the filter: %s", filters)
    func = partial(self.hosts_api.query_hidden_devices, filter=filters)
    id_list: List[str] = all_pages_numbered_offset_parallel(
        func=func,
        logger=self.logger,
        limit=SCROLL_BATCH_SIZE,
    )
    if not id_list:
        return DeviceNotFound

    return id_list


@filter_string
def hide(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict:
    """Hide a host or list of hosts within your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the device search.

    Returns
    -------
    dict: A dictionary containing details for the device hide result.
    """
    if not filters:
        raise MustProvideFilter

    return self._perform_action(
        action_name="hide_host",
        device_ids=self.get_device_ids(filters),
    )["resources"]


@filter_string
def unhide(
    self: HostsApiModule,
    filters: FalconFilter or str = None,
) -> Dict:
    """Unhide a host or list of hosts within your Falcon tenant.

    Arguments
    ---------
    filters: FalconFilter or str, optional
        Filters to apply to the device search.

    Returns
    -------
    dict: A dictionary containing details for the device unhide result.
    """
    if not filters:
        raise MustProvideFilter

    return self._perform_action(
        action_name="unhide_host",
        device_ids=self.get_hidden_ids(filters)
    )["resources"]



"""Caracara Hosts Module: host tagging functions.

In order to avoid the main hosts.py file getting too unwieldly, this file contains
the implementations of functions that can tag and untag hosts within the Falcon UI.
"""
# Disable the protected access rule because this file is an extension of the class in hosts.py.
# pylint: disable=protected-access
from __future__ import annotations
from typing import (
    Dict,
    List,
    TYPE_CHECKING,
)

from caracara.common.exceptions import (
    MustProvideFilter,
)
from caracara.filters import FalconFilter
from caracara.filters.decorators import filter_string

if TYPE_CHECKING:
    from caracara.modules.hosts import HostsApiModule


def _create_tag_list(potential_list: List[str] or str) -> List[str]:
    """Create a properly formatted list from a list, a string or a comma-delimited string."""
    tag_list = potential_list
    if not isinstance(potential_list, list):
        tag_list = potential_list.split(",")

    return tag_list


def _update_device_tags(
    self: HostsApiModule,
    action_name: str,
    device_ids: List[str],
    tag_list: List[str]
) -> Dict:
    """Tag or untag a device within the tenant."""
    return self.hosts_api.update_device_tags(
        action_name=action_name,
        ids=device_ids,
        tags=tag_list
    )["body"]


@filter_string
def tag(
    self: HostsApiModule,
    tags: List[str] or str, filters: FalconFilter or str = None,
) -> Dict:
    """Tag a host or list of hosts within your Falcon tenant.

    Arguments
    ---------
    tags: List of strings
        Tags to be applied to the discovered devices.
    filters: FalconFilter or str, optional
        Filters to apply to the device search.

    Returns
    -------
    dict: A dictionary containing details for the device tagging result.
    """
    if not filters:
        raise MustProvideFilter

    return self._update_device_tags(
        action_name="add",
        tag_list=self._create_tag_list(tags),
        device_ids=self.get_device_ids(filters)
    )["resources"]


@filter_string
def untag(
    self: HostsApiModule,
    tags: List[str] or str, filters: FalconFilter or str = None,
) -> Dict:
    """Untag a host or list of hosts within your Falcon tenant.

    Arguments
    ---------
    tags: List of strings
        Tags to be applied to the discovered devices.
    filters: FalconFilter or str, optional
        Filters to apply to the device search.

    Returns
    -------
    dict: A dictionary containing details for the device tagging result.
    """
    if not filters:
        raise MustProvideFilter

    return self._update_device_tags(
        action_name="remove",
        tag_list=self._create_tag_list(tags),
        device_ids=self.get_device_ids(filters)
    )["resources"]



"""Caracara HostsApiModule filters."""
import datetime

from caracara.common.constants import FILTER_OPERATORS, PLATFORMS
from caracara.filters.utils import (
    ISOTimestampChecker,
    RelativeTimestamp,
)
from caracara.filters.fql import FalconFilterAttribute


class HostContainedFilterAttribute(FalconFilterAttribute):
    """Filter hosts based on containment status."""

    name = "Contained"
    fql = "status"
    options = {
        "Contained": "contained",
        "Containment Pending": "containment_pending",
        "Not Contained": "normal",
    }
    restrict = True


class HostConnectionStatusFilterAttribute(FalconFilterAttribute):
    """
    Filter by whether a host is connected to the Falcon cloud.

    Current valid options are Online and Offline.
    """

    name = "ConnectionStatus"
    fql = "connection_status"
    options = ["Online", "Offline"]
    restrict = True

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter allows you to search for systems based on whether Falcon assesses them "
            "as being Online or Offline. Theoretically, all 'Online' systems should be available "
            "to connect to via Real Time Response (RTR)."
        )


class HostDomainFqlFilterAttribute(FalconFilterAttribute):
    """Filter by host AD domain."""

    name = "Domain"
    fql = "machine_domain"

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter accepts an AD domain, e.g. GOODDOMAIN or "
            "gooddomain.company.com. You can also provide multiple "
            "domains as a Python list or comma delimited string"
        )


class HostGroupIdFilterAttribute(FalconFilterAttribute):
    """Filter by host group ID."""

    name = "GroupID"
    fql = "groups"

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter accepts one or more Group IDs as either one "
            "string, or as a comma delimited list of strings. For example, "
            "075e03f5e5c04d83b4831374e7dc01c3 would target hosts within "
            "the group with ID 075e03f5e5c04d83b4831374e7dc01c3 only, or "
            "abcdefg123,abcdefg321 would target hosts in either group."
        )


class HostHostnameFilterAttribute(FalconFilterAttribute):
    """Filter by system hostname."""

    name = "Hostname"
    fql = "hostname"

    def example(self) -> str:
        """Show filter example."""
        return (
            "Provide either a single hostname string, or a list of hostnames "
            "via a comma delimited string or Python list. For example, "
            "you can omit two specific hosts with Hostname__NOT=HOST1,HOST2"
        )


class HostLastSeenFilterAttrribute(FalconFilterAttribute):
    """Filter for hosts last seen within a given timeframe."""

    name = "LastSeen"
    fql = "last_seen"
    types = [str]
    valid_operators = [
        "EQUAL",
        "GT",
        "GTE",
        "LT",
        "LTE",
    ]
    operator = "GTE"

    def _check_value(self, value) -> bool:
        """Confirm the provided value."""
        # Recycle the main check code first to make sure
        # we are working with a string input
        super()._check_value(value)

        if value.startswith(('-', '+')):
            rel_ts = RelativeTimestamp()
            returned = rel_ts.check_relative_timestamp(value)
        else:
            itc = ISOTimestampChecker()
            returned = itc.is_iso_format(value)

        return returned

    def get_fql(self) -> str:
        """Retrieve FQL syntax string."""
        if self.value is None:
            return ""

        operator_symbol = FILTER_OPERATORS[self.operator]

        if self.value.startswith(('-', '+')):
            rel_ts = RelativeTimestamp()
            new_timestamp = rel_ts.convert_relative_timestamp(
                datetime.datetime.utcnow(),
                self.value
            )
            value = new_timestamp.strftime(
                '%Y-%m-%dT%H:%M:%SZ'
            )
        else:
            value = self.value

        fql_string = f"{self.fql}: {operator_symbol}'{value}'"
        return fql_string

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter accepts two types of parameter: a fixed ISO 8601 "
            "timestamp (such as 2020-01-01:01:00:00Z), or a relative "
            "timestamp such as -30m. -30m means time now, minus thirty "
            "minutes, so is best combined with an operator such as GTE. "
            "A popular example is LastSeen__GTE=-30m, to stipulate all hosts "
            "that have been online in the past half hour (i.e. are likely to "
            "be online)"
        )


class HostFirstSeenFilterAttribute(HostLastSeenFilterAttrribute):
    """Filter for hosts first seen by Falcon within a given timeframe."""

    # NB: this filter inherits from Last Seen to avoid tonnes of duplicated code
    name = "FirstSeen"
    fql = "first_seen"

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter accepts two types of parameter: a fixed ISO 8601 "
            "timestamp (such as 2020-01-01:01:00:00Z), or a relative "
            "timestamp such as -30m. -30m means time now, minus thirty "
            "minutes, so is best combined with an operator such as GTE. "
            "One example is FirstSeen__GTE=-1d, to filter for all new hosts "
            "that have been added to Falcon within the past 1 day."
        )


class HostLocalIPFilterAttribute(FalconFilterAttribute):
    """Filter by host local IP addresses."""

    name = "LocalIP"
    fql = "local_ip"

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter accepts an IP address string associated with a "
            "network card, e.g. 172.16.1.2 or 172.16.* to cover the /16 "
            "range. You can also comma delimit strings for multiple matches, "
            "e.g. 172.16.1.2,172.16.1.3 to target hosts with each of those "
            "IPs, or provide a Python list of IP strings"
        )


class HostOSFqlFilterAttribute(FalconFilterAttribute):
    """
    Filter by host operating system types.

    Current valid options are Windows, Mac and Linux.
    """

    name = "OS"
    fql = "platform_name"
    options = PLATFORMS
    restrict = True


class HostOSVersionFqlFilterAttribute(FalconFilterAttribute):
    """Filter by Operating System version (e.g., Windows 7, RHEL 7.9, etc.)."""

    name = "OSVersion"
    fql = "os_version"

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter accepts a name of an operating system version "
            "and can be supplied many times. For example, Windows 7, "
            "RHEL 7.9, Catalina (10.15), etc."
        )


class HostRoleFilterAttribute(FalconFilterAttribute):
    """
    Filter by host role.

    Current valid options are DC, Server and Workstation.
    """

    name = "Role"
    fql = "product_type_desc"
    options = {
        "DC": "Domain Controller",
        "Server": "Server",
        "Workstation": "Workstation",
    }
    restrict = True


class HostSiteFilterAttribute(FalconFilterAttribute):
    """Filter by site name."""

    name = "Site"
    fql = "site_name"

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter accepts one or more site names as either one "
            "string, or as a comma delimtied list of strings. For example, "
            "London,Manchster1,Manchester2 would target hosts within any of "
            "those three sites."
        )


class HostTagFilterAttribute(FalconFilterAttribute):
    """
    Filter by sensor tag.

    - Tags created in Falcon should be prepended with FalconGroupingTag/
    - Tags created during the Sensor installion, or added to the restriry sould be prepended with
      SensorGroupingTag/
    """

    name = "Tag"
    description = "Filter by sensor tag"
    fql = "tags"

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter accepts one or more sensor tags as either one "
            "string, or as a comma delimited list of strings. For example, "
            "tag1,tag2,tag3 to filter by hosts with one of those tags."
        )


class HostOUFilterAttribute(FalconFilterAttribute):
    """Filter by Organisational Unit (OU)."""

    name = "OU"
    fql = "ou"

    def example(self) -> str:
        """Show filter example."""
        return (
            "This filter accepts an Organisational Unit (OU) name as a "
            "string. You can also comma delimit OUs for multiple matches, "
            "e.g. UKServers,USServers to target hosts within any of those "
            "OUs. Programmatically, you can pass a Python list of OUs."
        )


FILTER_ATTRIBUTES = [
    HostContainedFilterAttribute,
    HostConnectionStatusFilterAttribute,
    HostDomainFqlFilterAttribute,
    HostGroupIdFilterAttribute,
    HostHostnameFilterAttribute,
    HostLastSeenFilterAttrribute,
    HostFirstSeenFilterAttribute,
    HostLocalIPFilterAttribute,
    HostOSFqlFilterAttribute,
    HostOSVersionFqlFilterAttribute,
    HostRoleFilterAttribute,
    HostSiteFilterAttribute,
    HostTagFilterAttribute,
    HostOUFilterAttribute,
]



"""Real Time Response (RTR) Constants."""
RTR_COMMANDS = {
    "cat": "read_only",
    "cd": "read_only",
    "clear": "read_only",
    "cp": "active_responder",
    "csrutil": "read_only",
    "cswindiag": "admin",
    "encrypt": "active_responder",
    "env": "read_only",
    "eventlog": "read_only",
    "filehash": "read_only",
    "get": "active_responder",
    "getsid": "read_only",
    "history": "read_only",
    "ifconfig": "read_only",
    "ipconfig": "read_only",
    "kill": "active_responder",
    "ls": "read_only",
    "map": "active_responder",
    "memdump": "active_responder",
    "mkdir": "active_responder",
    "mount": "read_only",
    "mv": "active_responder",
    "netstat": "read_only",
    "ps": "read_only",
    "put": "admin",
    "put-and-run": "admin",
    "reg": {
        "_default": "active_responder",
        "query": "read_only",
    },
    "restart": "active_responder",
    "rm": "active_responder",
    "run": "admin",
    "runscript": {
        "_default": "admin",
        "-Raw": "admin",
        "-CloudFile": "active_responder",
        "-HostPath": "admin",
    },
    "shutdown": "active_responder",
    "tar": "active_responder",
    "umount": "active_responder",
    "unmap": "active_responder",
    "update": "active_responder",
    "users": "read_only",
    "xmemdump": "active_responder",
    "zip": "active_responder",
}

# Maximum number of devices within an RTR batch session
# There is a 10,000 limit enforced in EU-1 and US-2, so we use this as our limit
MAX_BATCH_SESSION_HOSTS = 10000

# Most number of threads permitted when running RTR batch commands
MAX_BATCH_SESSION_THREADS = 3

# RTR sessions last for 10 minutes
SESSION_EXPIRY = 60 * 10

# Refresh RTR sessions if there are < 3 minutes left
SESSION_REFRESH_TIMEOUT = 60 * 3
